package jp.hishidama.swing.table;

import java.util.Vector;

import jp.hishidama.swing.table.LazyTable.LazyAddAction;

/**
 * xXVJTablepf[^f.
 *
 * @author <a target="hishidama"
 *         href="http://www.ne.jp/asahi/hishidama/home/tech/java/swing/JTable.html"
 *         >Ђ</a>
 * @see LazyTable
 * @since 2009.03.29
 * @version 2009.04.26 ExTableModelh悤ύX
 */
public class LazyTableModel extends ExTableModel {
	private static final long serialVersionUID = 1591234686660546702L;

	@Override
	public void setValueAt(Object value, int row, int column) {
		if (value instanceof ChangeValue) {
			super.setValueAt(value, row, column);
			return;
		}
		Object old = super.getValueAt(row, column);
		if (old instanceof ChangeValue) {
			ChangeValue cv = ((ChangeValue) old).clone();
			cv.setValue(value);
			value = cv;
		} else {
			value = new ChangeValue(LazyTable.CHG, value, old);
		}
		super.setValueAt(value, row, column);
	}

	/**
	 * l擾.
	 * <p>
	 * ύX̏ꍇ́A̐VlB
	 * </p>
	 */
	@Override
	public Object getValueAt(int row, int column) {
		Object value = super.getValueAt(row, column);
		if (value instanceof ChangeValue) {
			ChangeValue cv = (ChangeValue) value;
			return cv.getValue();
		} else {
			return value;
		}
	}

	/**
	 * Âl擾.
	 *
	 * @param row
	 * @param column
	 * @return ύX̏ꍇ́ǍÂlBύXłȂꍇ{@link #getValueAt(int, int)}ƓB
	 */
	public Object getOldValueAt(int row, int column) {
		Object value = super.getValueAt(row, column);
		if (value instanceof ChangeValue) {
			ChangeValue cv = (ChangeValue) value;
			return cv.getOldValue();
		} else {
			return value;
		}
	}

	/**
	 * ftHgl擾.
	 * <p>
	 * I[o[ChĎgpÕ\bhB<br>
	 * sǉۂɓ\bhĂ΂AlƂĐݒ肳B
	 * </p>
	 *
	 * @param row
	 * @param column
	 * @return ftHgl
	 * @see LazyAddAction
	 */
	public Object getDefaultValueAt(int row, int column) {
		return null;
	}

	/**
	 * sXe[^X擾.
	 * <p>
	 * w肳ꂽs璲AŏɕύXiǉE폜jƂɂ̃Xe[^XԂB
	 * </p>
	 *
	 * @param row
	 *            sCfbNX
	 * @return Xe[^Xi{@link LazyTable#CHG}j
	 */
	@SuppressWarnings("unchecked")
	public int getRowStatus(int row) {
		Vector rowVector = (Vector) dataVector.elementAt(row);
		for (Object v : rowVector) {
			if (v instanceof ChangeValue) {
				ChangeValue cv = (ChangeValue) v;
				return cv.getType();
			}
		}
		return 0;
	}

	/**
	 * Xe[^X擾.
	 *
	 * @param row
	 *            sCfbNX
	 * @param column
	 *            CfbNX
	 * @return Xe[^Xi{@link LazyTable#CHG}j
	 */
	public int getStatusAt(int row, int column) {
		Object value = super.getValueAt(row, column);
		if (value instanceof ChangeValue) {
			ChangeValue cv = (ChangeValue) value;
			return cv.getType();
		} else {
			return 0;
		}
	}

	/**
	 * Xe[^Xݒ.
	 *
	 * @param row
	 *            sCfbNX
	 * @param type
	 *            Xe[^Xi{@link LazyTable#CHG}j
	 */
	@SuppressWarnings("unchecked")
	public void setRowStatus(int row, int type) {
		Vector rowVector = (Vector) dataVector.elementAt(row);
		for (int i = 0; i < rowVector.size(); i++) {
			Object v = rowVector.get(i);
			if (v instanceof ChangeValue) {
				ChangeValue cv = (ChangeValue) v;
				cv.setType(type);
			} else {
				ChangeValue cv = new ChangeValue(type, v, v);
				super.setValueAt(cv, row, i);
			}
		}
	}

	/**
	 * ύXm.
	 */
	@SuppressWarnings("unchecked")
	public void commit() {
		loop: for (int i = getRowCount() - 1; i >= 0; i--) {
			Vector rowVector = (Vector) dataVector.elementAt(i);
			for (Object v : rowVector) {
				if (v instanceof ChangeValue) {
					ChangeValue cv = (ChangeValue) v;
					switch (cv.getType()) {
					case LazyTable.ADD:
					case LazyTable.CHG:
						changeNew(rowVector, i);
						continue loop;
					case LazyTable.DEL:
						changeDel(rowVector, i);
						continue loop;
					}
				}
			}
		}
		undoSupport.discardAllEdits();
	}

	/**
	 * ύX߂.
	 */
	@SuppressWarnings("unchecked")
	public void rollback() {
		loop: for (int i = getRowCount() - 1; i >= 0; i--) {
			Vector rowVector = (Vector) dataVector.elementAt(i);
			for (Object v : rowVector) {
				if (v instanceof ChangeValue) {
					ChangeValue cv = (ChangeValue) v;
					switch (cv.getType()) {
					case LazyTable.ADD:
						changeDel(rowVector, i);
						continue loop;
					case LazyTable.CHG:
					case LazyTable.DEL:
						changeOld(rowVector, i);
						continue loop;
					}
				}
			}
		}
		undoSupport.discardAllEdits();
	}

	protected void changeNew(Vector<Object> rowVector, int row) {
		int sz = rowVector.size();
		for (int i = 0; i < sz; i++) {
			Object v = rowVector.get(i);
			if (v instanceof ChangeValue) {
				ChangeValue cv = (ChangeValue) v;
				rowVector.set(i, cv.getValue());
			}
		}
		fireTableRowsUpdated(row, row);
	}

	protected void changeOld(Vector<Object> rowVector, int row) {
		int sz = rowVector.size();
		for (int i = 0; i < sz; i++) {
			Object v = rowVector.get(i);
			if (v instanceof ChangeValue) {
				ChangeValue cv = (ChangeValue) v;
				rowVector.set(i, cv.getOldValue());
			}
		}
		fireTableRowsUpdated(row, row);
	}

	protected void changeDel(Vector<Object> rowVector, int row) {
		removeRow(row);
	}

	/**
	 * ύXO̒lێB
	 */
	public static class ChangeValue implements Cloneable {

		protected int type;
		protected Object value, old;

		/**
		 * RXgN^[.
		 *
		 * @param type
		 *            Xe[^X
		 * @param value
		 *            Vl
		 * @param old
		 *            Âl
		 */
		public ChangeValue(int type, Object value, Object old) {
			this.type = type;
			this.value = value;
			this.old = old;
		}

		/**
		 * Xe[^Xݒ.
		 *
		 * @param type
		 *            Xe[^X
		 */
		public void setType(int type) {
			this.type = type;
		}

		/**
		 * Xe[^X擾.
		 *
		 * @return Xe[^X
		 */
		public int getType() {
			if ((type & LazyTable.CHG) != 0) {
				int t = type;
				if (value == null) {
					if (old == null) {
						t &= ~LazyTable.CHG;
						return t;
					}
				} else {
					if (value.equals(old)) {
						t &= ~LazyTable.CHG;
						return t;
					}
				}
			}
			return type;
		}

		/**
		 * Vlݒ.
		 *
		 * @param value
		 *            l
		 */
		public void setValue(Object value) {
			this.value = value;
		}

		/**
		 * Vl擾.
		 *
		 * @return l
		 */
		public Object getValue() {
			return value;
		}

		/**
		 * Âlݒ.
		 *
		 * @param old
		 *            l
		 */
		public void setOldValue(Object old) {
			this.old = old;
		}

		/**
		 * Âl擾.
		 *
		 * @return l
		 */
		public Object getOldValue() {
			return old;
		}

		@Override
		public String toString() {
			return String.valueOf(value);
		}

		@Override
		protected ChangeValue clone() {
			try {
				return (ChangeValue) super.clone();
			} catch (CloneNotSupportedException e) {
				throw new InternalError(e.toString());
			}
		}
	}

	@Override
	protected Vector<?> cloneRowData(Vector<?> rowData) {
		@SuppressWarnings("unchecked")
		Vector<Object> v = (Vector<Object>) super.cloneRowData(rowData);
		for (int i = 0; i < v.size(); i++) {
			Object obj = v.get(i);
			if (obj instanceof ChangeValue) {
				ChangeValue cv = (ChangeValue) obj;
				v.set(i, cv.clone());
			}
		}
		return v;
	}
}
